from __future__ import annotations

import argparse
import logging
from dataclasses import dataclass

from clippy.cookbook import Cookbook

from rover.netbox import Netbox

FIND_VM_QUERY = """
query find_vm($name: [String!]) {
    virtual_machine_list(filters: {name: {in_list: $name}}) {
        id
        name

        primary_ip4 {
            dns_name
        }

        interfaces {
            ip_addresses {
                id
                role
                address
                dns_name
            }
        }
    }
}
"""


@dataclass
class VmToRemove:
    netbox_id: int
    fqdn: str
    ips_to_delete: list[int]


class RemoveVm(Cookbook):
    """Shuts down a virtual machine."""

    def argument_parser(self) -> argparse.ArgumentParser:
        parser = super().argument_parser()
        parser.add_argument("fqdn")
        parser.add_argument("--skip-shutdown", help="skip shutting down the VM", action="store_true")
        return parser

    def _find_vm(self, netbox: Netbox) -> VmToRemove:
        self.logger.debug("Querying Netbox for '%s'", self.args.fqdn)
        search_for = [self.args.fqdn]
        if "." in self.args.fqdn:
            search_for.append(self.args.fqdn.split(".")[0])

        results = netbox.query(gql=FIND_VM_QUERY, op="find_vm", variables={"name": search_for})["virtual_machine_list"]
        if len(results) != 1:
            raise ValueError(f"No VMs found for '{self.args.fqdn}")
        vm = results[0]

        if not vm["primary_ip4"]["dns_name"]:
            raise ValueError(f"VM '{self.args.fqdn}' does not have a DNS name assigned")

        return VmToRemove(
            netbox_id=int(vm["id"]),
            fqdn=vm["primary_ip4"]["dns_name"],
            ips_to_delete=[
                int(ip["id"])
                for interface in vm["interfaces"]
                for ip in interface["ip_addresses"]
                if ip.get("role", "") == ""
            ],
        )

    def execute(self):
        netbox = Netbox.from_config(self.clippy.config("netbox"))
        logging.getLogger("rover").setLevel(logging.getLogger("clippy").getEffectiveLevel())

        vm = self._find_vm(netbox)
        self.logger.info("Found VM %s (%s)", vm.fqdn, vm.netbox_id)

        if self.args.skip_shutdown:
            self.logger.warning("Skipping shutdown as requested")
        else:
            self.logger.info("Shutting down VM")
            remote = self.clippy.remote_hosts(f"D{{{vm.fqdn}}}")
            remote.execute(command="nohup shutdown now &> /dev/null & exit", show_progress=False, show_output=False)

        self.logger.info("Removing VM from Puppet")
        self.clippy.puppet_hosts(vm.fqdn).clean()

        self.logger.info("Updating Netbox")
        netbox.update_object(f"virtualization/virtual-machines/{vm.netbox_id}", {"status": "offline"})

        for ip in vm.ips_to_delete:
            netbox.update_object(f"ipam/ip-addresses/{ip}", {"dns_name": ""})

        self.logger.info("Updating DNS data")
        dnsgen = self.clippy.remote.query("P{P:dns::dnsgen}", single=True)
        dnsgen.execute(command="update-dns", show_output=True)
