import logging
from typing import Any

import requests
from clippy.exceptions import AutomationError
from python_graphql_client import GraphqlClient

from rover.os import DebianVersion

LOGGER = logging.getLogger(__name__)


class NetboxError(AutomationError):
    pass


GET_PLATFORMS_QUERY = """
query platforms {
    platform_list {
        id
        name
        custom_fields
    }
}
"""


class Netbox:
    def __init__(self, endpoint: str, token: str) -> None:
        self._endpoint = endpoint.removesuffix("/")
        self._token = token

        self._session = requests.Session()
        self._session.headers.update({"Authorization": f"Token {token}"})

        self._gql = GraphqlClient(
            endpoint=f"{self._endpoint}/graphql/",
            headers={
                "Authorization": f"Token {token}",
                # why is this needed???
                "Content-Type": "application/json",
            },
        )

        self._platforms = None

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "Netbox":
        return cls(
            endpoint=config["endpoint"],
            token=config["token"],
        )

    def query(self, *, gql: str, op: str, variables: dict[str, Any]) -> dict[str, Any]:
        try:
            return self._gql.execute(query=gql, operation_name=op, variables=variables)["data"]
        except requests.exceptions.HTTPError as e:
            raise AutomationError(e.response.text) from e

    def get_platform_id(self, version: DebianVersion) -> int:
        if not self._platforms:
            LOGGER.info("Loading platform data from Netbox")
            self._platforms = {
                DebianVersion(result["custom_fields"]["debian_codename"]): int(result["id"])
                for result in self.query(gql=GET_PLATFORMS_QUERY, op="platforms", variables={})["platform_list"]
                if result["custom_fields"]["debian_codename"] is not None
            }
            LOGGER.debug("Platforms: %s", self._platforms)

        return self._platforms[version]

    def get(self, path: str):
        response = self._session.get(f"{self._endpoint}/api/{path}/")
        response.raise_for_status()
        return response.json()

    def create_object(self, path: str, data: dict[str, Any]) -> dict[str, Any]:
        response = self._session.post(f"{self._endpoint}/api/{path}/", json=data)
        response.raise_for_status()
        return response.json()

    def update_object(self, path: str, data: dict[str, Any]) -> dict[str, Any]:
        response = self._session.patch(f"{self._endpoint}/api/{path}/", json=data)
        response.raise_for_status()
        return response.json()
