import time
from abc import ABC, abstractmethod
from dataclasses import dataclass

from clippy.exceptions import AutomationError
from clippy.remote import Remote


class VirtError(AutomationError):
    pass


class VirtCluster(ABC):
    netbox_id: int
    label: str

    @abstractmethod
    def create_vm(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> None:
        pass

    @abstractmethod
    def is_running(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> bool:
        pass

    @abstractmethod
    def start_vm(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> None:
        pass


@dataclass
class Hypervisor:
    netbox_id: int
    fqdn: str


@dataclass
class VirtualMachine:
    netbox_id: int
    fqdn: str


class ProxmoxCluster(VirtCluster):
    def __init__(
        self,
        remote: Remote,
        netbox_id: int,
        label: str,
    ):
        self._remote = remote
        self.netbox_id = netbox_id
        self.label = label

        self._id_cache: dict[tuple[str, str], int] = {}

    def _get_vm_id(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> int:
        cache_key = (hypervisor.fqdn, vm.fqdn)
        if cache_key in self._id_cache:
            return self._id_cache[cache_key]

        hostname = vm.fqdn.split(".")[0]
        result = (
            self._remote.query(f"D{{{hypervisor.fqdn}}}")
            .execute(command=f"qm list | grep {hostname} | awk '{{print $1}}'")
            .results
        )

        for nodeset, output in result:
            vm_id = int(output.message().decode().strip().split()[0])
            self._id_cache[cache_key] = vm_id
            return vm_id

        raise VirtError(f"Unable to determine virtual machine ID for {vm.fqdn}")

    def is_running(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> bool:
        vm_id = self._get_vm_id(hypervisor, vm)
        result = self._remote.query(f"D{{{hypervisor.fqdn}}}").execute(command=f"qm status {vm_id}").results
        for nodeset, output in result:
            return "running" in output.message().decode()
        raise VirtError(f"Unable to determine virtual machine status for {vm.fqdn}")

    def create_vm(self, hypervisor: Hypervisor, vm: VirtualMachine) -> None:
        self._remote.query(f"D{{{hypervisor.fqdn}}}").execute(
            command=f"/usr/local/sbin/majava-create-netbox-vm {vm.netbox_id}",
            show_output=True,
        )

    def start_vm(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> None:
        vm_id = self._get_vm_id(hypervisor, vm)
        self._remote.query(f"D{{{hypervisor.fqdn}}}").execute(command=f"qm start {vm_id}")
        time.sleep(10)


class GanetiCluster(VirtCluster):
    def __init__(
        self,
        remote: Remote,
        netbox_id: int,
        label: str,
    ):
        self._remote = remote
        self.netbox_id = netbox_id
        self.label = label

    def create_vm(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> None:
        raise RuntimeError("Not implemented yet")

    def is_running(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> bool:
        raise RuntimeError("Not implemented yet")

    def start_vm(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> None:
        raise RuntimeError("Not implemented yet")
